#
# ThumbscrewAppDelegate.py
# Thumbscrew
#
# Created by Zachery Bir on Fri Oct 15 2004.
# Copyright (c) 2004 Zachery bir. All rights reserved.
#

# library imports
import dircache
import sys
import os
import os.path
import md5
import re
import time
import new, objc
import datetime
from urllib import quote
import xml.etree.ElementTree as etree
from PyObjCTools import NibClassBuilder, AppHelper

# import needed classes/functions from Foundation
from Foundation import *

# import Nib loading functionality from AppKit
from AppKit import *

# local product imports
import PreferenceController
import OutputController
from constants import *

#EXPIRATION_DATE = datetime.datetime(2005, 11, 04, 12, 00)
EXPIRATION_DATE = None
TSPNGPboardType = u"Apple PNG pasteboard type"

def _hasGoodTypes(pboard, types):
    test_types = pboard.types()
    for pbtype in types:
        if test_types.containsObject_(pbtype):
            return True
    return False

def collect_filenames(directories):
    all_files = []
    for d in directories:
        for root, dirs, files in os.walk(d):
            all_files.extend([os.path.join(root, x) for x in files])
    return all_files

def ensurePathExists(path):
    try:
        os.makedirs(path)
    except os.error:
        pass

def thumbscrewPictureFolder():
    return os.path.join(NSHomeDirectory(),
                        u'Pictures',
                        u'Thumbscrew')

def thumbscrewBatchFolder():
    return os.path.join(NSHomeDirectory(),
                        u'Library',
                        u'Application Support',
                        u'Thumbscrew',
                        u'Batches')

untitled_number = re.compile(r'^untitled(\d*)\.')

def nextUntitledPath(suffix):
    untitleds = [-1]
    ensurePathExists(thumbscrewPictureFolder())
    for f in dircache.listdir(thumbscrewPictureFolder()):
        m = untitled_number.match(f.lower())
        if m:
            try:
                num = int(m.groups()[0])
            except ValueError:
                num = 0
            untitleds.append(num)
    return os.path.join(thumbscrewPictureFolder(),
                        u'Untitled%d.%s' % (max(untitleds) + 1, suffix))

def loadFrameworkFromBundleOrSystem(name):
    _mod = new.module(name)
    try:
        objc.loadBundle(
            name,
            _mod.__dict__,
            bundle_path="%s/%s" % (NSBundle.mainBundle().bundlePath(),
                                   "Contents/Frameworks/%s.framework" % name))
    except ImportError: # we were built with --alias, perhaps
        objc.loadBundle(
            name,
            _mod.__dict__,
            bundle_path="/Library/Frameworks/%s.framework" % name)
    return _mod

Growl = loadFrameworkFromBundleOrSystem('Growl')
Sparkle = loadFrameworkFromBundleOrSystem('Sparkle')

NibClassBuilder.extractClasses(u'MainMenu')

class DropView(NibClassBuilder.AutoBaseClass):
    """ A view for managing our drag and drop functionality
    """

    can_accept_drags = False

    def draggingEntered_(self, sender):
        if self.can_accept_drags:
            sourceDragMask = sender.draggingSourceOperationMask()
            pboard = sender.draggingPasteboard()

            if _hasGoodTypes(pboard,
                             [NSFilenamesPboardType, NSURLPboardType]):
                if sourceDragMask:
                    return NSDragOperationCopy

        return NSDragOperationNone

    def performDragOperation_(self, sender):
        if self.can_accept_drags:
            sourceDragMask = sender.draggingSourceOperationMask()
            pboard = sender.draggingPasteboard()

            if _hasGoodTypes(pboard, [NSFilenamesPboardType]):
                files = list(pboard.propertyListForType_(NSFilenamesPboardType))
                if sourceDragMask & NSDragOperationCopy:
                    self.app.application_openFiles_(self.app, files)
                    return True
                if _hasGoodTypes(pboard, [NSURLPboardType]):
                    url = list(pboard.propertyListForType_(NSURLPboardType))[0]
                    if sourceDragMask & NSDragOperationCopy:
                        data = NSData.alloc().initWithContentsOfURL_(
                            NSURL.URLWithString_(url))
                        data_name = os.path.join(u'/tmp', url.split('/')[-1])
                        data.writeToFile_atomically_(data_name, True)
                        self.app.application_openFiles_(self.app, [data_name])
                        return True
        return True

    def concludeDragOperation_(self, sender):
        pass

    def resetImage(self):
        start_image = NSImage.alloc().initWithContentsOfFile_(
            u'Thumbscrew_dropzone.png')
        self.setImage_(start_image)

    def paste_(self, sender):
        pboard = NSPasteboard.generalPasteboard()
        if _hasGoodTypes(pboard, [NSTIFFPboardType]):
            data = NSData.alloc().initWithData_(pboard.dataForType_(NSTIFFPboardType))
            data_name = nextUntitledPath('jpg')
            data.writeToFile_atomically_(data_name, True)
            return self.app.application_openFiles_(self.app, [data_name])
        if _hasGoodTypes(pboard, [TSPNGPboardType]):
            data = NSData.alloc().initWithData_(pboard.dataForType_(TSPNGPboardType))
            data_name = nextUntitledPath('png')
            data.writeToFile_atomically_(data_name, True)
            return self.app.application_openFiles_(self.app, [data_name])
        if _hasGoodTypes(pboard, [NSPICTPboardType]):
            data = NSData.alloc().initWithData_(pboard.dataForType_(NSPICTPboardType))
            data_name = nextUntitledPath('pict')
            data.writeToFile_atomically_(data_name, True)
            return self.app.application_openFiles_(self.app, [data_name])
        if _hasGoodTypes(pboard, [NSFilenamesPboardType]):
            files = list(pboard.propertyListForType_(NSFilenamesPboardType))
            return self.app.application_openFiles_(self.app, files)
        if _hasGoodTypes(pboard, [NSURLPboardType]):
            url = list(pboard.propertyListForType_(NSURLPboardType))[0]
            data = NSData.alloc().initWithContentsOfURL_(NSURL.URLWithString_(url))
            data_name = os.path.join(thumbscrewPictureFolder(),
                                     url.split('/')[-1])
            data.writeToFile_atomically_(data_name, True)
            return self.app.application_openFiles_(self.app, [data_name])
        if _hasGoodTypes(pboard, [NSStringPboardType]):
            p_files = pboard.stringForType_(NSStringPboardType).split()
            return self.app.application_openFiles_(
                self.app, [p for p in p_files if os.path.exists(p)])
        return True


class ThumbscrewAppDelegate(NibClassBuilder.AutoBaseClass):
    """
    The controller for Thumbscrew
    """

    def performClose_(self, sender):
        keyWindow = NSApplication.sharedApplication().keyWindow()
        if keyWindow is self.window:
            return False
        keyWindow.performClose_(sender)

    def validateMenuItem_(self, menuItem):
        keyWindow = NSApplication.sharedApplication().keyWindow()
        if keyWindow is self.window and menuItem.title() == u'Save':
            return False
        if keyWindow is not self.window and menuItem.title() == u'Paste':
            return False
        return True

    def paste_(self, sender):
        self.drop_zone.paste_(sender)

    def saveOutput_(self, sender):
        keyWindow = NSApplication.sharedApplication().keyWindow()
        if (self.activity_output_controller and
            keyWindow is self.activity_output_controller.window()):
            return self.activity_output_controller.saveOutput()
        return False

    def showPreferencePanel_(self, sender):
        if self.preference_controller is None:
            self.preference_controller = PreferenceController.PreferenceController.alloc().init(self)
        self.preference_controller.showWindow_(self)

    def showMainWindow_(self, sender):
        self.window.makeKeyAndOrderFront_(sender)

    def showActivityOutput_(self, sender):
        activity_output = u'\n'.join(self.activity_output)
        if self.activity_output_controller is None:
            self.activity_output_controller = OutputController.OutputController.alloc().init(u'Activity Log', activity_output, editable=False)
        else:
            self.activity_output_controller.setData(activity_output)
        self.activity_output_controller.showWindow_(self)

    def showHelp_(self, sender):
        NSWorkspace.sharedWorkspace().openFile_(
            NSBundle.mainBundle().pathForResource_ofType_(u'Thumbscrew Manual', u'pdf'))

    def application_openFiles_(self, app, filenames):
        """
        Invoked by NSApplication when the app attempts to open multiple files.
        """
        valid_img_exts = [u'.jpg', u'.jpeg',
                          u'.gif',
                          u'.tif', u'.tiff',
                          u'.png',
                          u'.pict',
                          u'.raw', u'.nef',
                          ]

        def checkext(filename):
            ext = unicode(os.path.splitext(filename)[1].lower())
            return ext in valid_img_exts

        some_dirs = []
        for fname in filenames:
            if os.path.isdir(fname):
                some_dirs.append(fname)
                filenames.remove(fname)
        if some_dirs:
            if (self._recursively_process_folders or 
                NSRunAlertPanel(u"Process Folder Contents?",
                                u"You passed in folders. Do you want to"
                                u" recursively process all images inside?",
                                u"Yes", u"Cancel", None)):
                filenames = filenames + collect_filenames(some_dirs)
            else:
                return

        self.start_time = datetime.datetime.fromtimestamp(time.time())
        self.activity_output.append(u"Starting at: %s" %
                                    self.start_time.strftime('%r'))
        self.files = [x for x in filenames if checkext(x)]

        if self.subprocsReady:
            self.sendBatch()
        else:
            self.queue = self.files[:]

        return

    def sendBatch(self):
        dc = NSDistributedNotificationCenter.defaultCenter()
        dc.postNotificationName_object_userInfo_(
            kThumbnailerActivity,
            kThumbnailReceiveFilenamesEvent,
            NSDictionary.dictionaryWithDictionary_(
            {u'filenames': self.files}))
        if self._resize_origs:
            dc.postNotificationName_object_userInfo_(
                kResizerActivity,
                kThumbnailReceiveFilenamesEvent,
                NSDictionary.dictionaryWithDictionary_(
                {u'filenames': self.files}))

    def postProcessFilenames(self):
        now = datetime.datetime.fromtimestamp(time.time())
        batch_name = now.strftime('%Y%m%d%H%M%S')

        thumbnail_dir = self._save_thumbnails_path or None
        resized_dir = self._save_resize_origs_path or None

        root = etree.Element(u'thumbscrew_batch')
        root.set(u'name', batch_name)

        for f in self.files:
            original_dir, filename_extension = os.path.split(f)
            filename, extension = os.path.splitext(filename_extension)

            node = etree.SubElement(root, u'image')
            node.set(u'name', filename)

            orig = etree.SubElement(node, u'version')
            orig.set(u'name', u'original')
            orig.set(u'path', f)

            thumbnail_path = os.path.join(
                (thumbnail_dir or original_dir),
                u'%s%s.%s' % (filename,
                              self._thumbnail_suffix,
                              self._thumbnail_image_format.lower()))
            thumbnail = etree.SubElement(node, u'version')
            thumbnail.set(u'name', u'thumbnail')
            thumbnail.set(u'path', thumbnail_path)

            if self._resize_origs:
                resized_path = os.path.join(
                    (resized_dir or original_dir),
                    u'%s%s%s' % (filename,
                                 self._resize_suffix,
                                 extension))
                resized = etree.SubElement(node, u'version')
                resized.set(u'name', u'resized')
                resized.set(u'path', resized_path)

        tree = etree.ElementTree(root)
        ensurePathExists(thumbscrewBatchFolder())
        batch_file = os.path.join(thumbscrewBatchFolder(),
                                  '%s.xml' % batch_name)
        tree.write(batch_file)

        if self._post_flight_script:
            batch_process = NSTask.alloc().init()
            batch_process.setLaunchPath_(self._post_flight_script)
            batch_process.setArguments_([batch_file])
            batch_environ = os.environ.copy()
            del batch_environ['PYTHONPATH']
            batch_process.setEnvironment_(batch_environ)
            batch_process.launch()

            Growl.GrowlApplicationBridge.notifyWithTitle_description_notificationName_iconData_priority_isSticky_clickContext_(
                self._post_flight_script,
                u'Processed the post flight script.',
                u'Post flight script complete',
                None,
                0,
                False,
                None)

    def markProgress(self, value, done=False, label=None):
        if label:
            self.progress_name.setStringValue_(label)
        self.progress_bar.incrementBy_(value)
        self.progress_name.displayIfNeeded()
        self.progress_bar.displayIfNeeded()
        if done:
            self.progress_name.setStringValue_(u'')
            self.progress_name.displayIfNeeded()
            self.progress_bar.setDoubleValue_(0.0)
            self.progress_bar.stopAnimation_(self)

    def markProgressDone_(self, notification):
        info = notification.userInfo() or {}
        filename = info.get(u'filename', None)
        path, fn = os.path.split(filename)
        self.activity_output.append(u"Created: %s" % filename)
        if filename:
            self.drop_zone.setImage_(
                NSImage.alloc().initWithContentsOfFile_(filename))
        self.drop_zone.displayIfNeeded()
        self.markProgress(100.0 / len(self.files))
        Growl.GrowlApplicationBridge.notifyWithTitle_description_notificationName_iconData_priority_isSticky_clickContext_(
            u'Thumbscrewed',
            fn,
            u'Thumbnail Done',
            NSData.dataWithContentsOfFile_(filename),
            0,
            False,
            None)

    def markProgressLabel_(self, notification):
        info = notification.userInfo() or {}
        filename = info.get(u'filename', None)
        self.activity_output.append(u"Thumbnailing: %s" % filename)
        if filename:
            self.drop_zone.setImage_(
                NSImage.alloc().initWithContentsOfFile_(filename))
            self.drop_zone.displayIfNeeded()
            label = (u"%d/%d - %s" % (self.files.index(filename) + 1,
                                      len(self.files),
                                      os.path.split(filename)[1]))
            self.markProgress(1.0, label=label)

    def cancelProcessing_(self, sender):
        self.cancelThumbnailer_(sender)
        self.cancelResizer_(sender)
        self.files = []

    def cancelThumbnailer_(self, notification):
        if self.thumbnailer:
            self.thumbnailer.terminate()
            self.thumbnailer = None
        self.drop_zone.resetImage()

    def cancelResizer_(self, notification):
        if self.resizer:
            self.resizer.terminate()
            self.resizer = None
        self.drop_zone.resetImage()

    def endOfBatch_(self, notification):
        self.markProgress(100.0, done=True)
        self.setUpDropZone_(notification)
        self.postProcessFilenames()
        end_time = datetime.datetime.fromtimestamp(time.time())
        total_time = end_time - self.start_time
        self.activity_output.append(u"Finished at: %s" % end_time.strftime('%r'))
        self.activity_output.append(u"Total time: %s" % total_time)
        Growl.GrowlApplicationBridge.notifyWithTitle_description_notificationName_iconData_priority_isSticky_clickContext_(
            u'Batch Done',
            u'%d images processed in %d second%s' % (len(self.files),
                                                     total_time.seconds,
                                                     total_time.seconds != 1 and u's' or u''),
            u'Batch Done',
            None,
            0,
            False,
            None)

    def awakeFromNib(self):
        #self.window.setBackgroundColor_(
        #    NSColor.colorWithPatternImage_(
        #        NSImage.imageNamed_(u"background.tif")))

        self.drop_zone.app = self
        self.drop_zone.resetImage()
        self.progress_bar.setIndeterminate_(True)
        self.progress_bar.startAnimation_(self)
        self.progress_name.setStringValue_(u'Talking to Thumbnailer...')

        pinger = NSDistributedNotificationCenter.defaultCenter()

        regs = (
  (self.setUpDropZone_, kThumbnailerActivity, kPreferencesSetEvent),
  (self.markProgressLabel_, kThumbnailerActivity, kThumbnailStartingEvent),
  (self.markProgressDone_, kThumbnailerActivity, kThumbnailDoneEvent),
  (self.cancelThumbnailer_, kThumbnailerActivity, kThumbnailProcessDoneEvent),
  (self.endOfBatch_, kThumbnailerActivity, kThumbnailBatchDoneEvent),
  (self.cancelResizer_, kResizerActivity, kThumbnailProcessDoneEvent),
  (self.reloadPrefs_, kThumbnailerActivity, kReadyForPreferencesEvent),
  (self.reloadPrefs_, kResizerActivity, kReadyForPreferencesEvent)
            )
        for reg in regs:
            pinger.addObserver_selector_name_object_(self, *reg)

    def applicationWillFinishLaunching_(self, notification):
        # Check for out of date
        if EXPIRATION_DATE is not None:
            if EXPIRATION_DATE < datetime.datetime.fromtimestamp(time.time()):
                if NSRunAlertPanel(u"Beta Expired.",
                                   u"This version of Thumbscrew is expired."
                                   u" Please check http://www.urbanape.com/"
                                   u" for the latest version",
                                   u"Quit", None, None):
                    return NSApplication.sharedApplication().terminate_(self)
        # We'll kick of both subprocesses - they'll both die when
        # we're done, and it may speed things up later on.
        self.thumbnailer = NSTask.launchedTaskWithLaunchPath_arguments_(
            NSBundle.mainBundle().pathForResource_ofType_(u'thumbnailer', None),
              [kThumbnailerActivity])

        self.resizer = NSTask.launchedTaskWithLaunchPath_arguments_(
            NSBundle.mainBundle().pathForResource_ofType_(u'thumbnailer', None),
              [kResizerActivity])

    def applicationWillTerminate_(self, notification):
        self.cancelProcessing_(self)

    def registrationDictionaryForGrowl(self):
        return { u'AllNotifications' : [ u'Thumbnail Done',
                                         u'Post flight script complete',
                                         u'Batch Done' ],
                 u'DefaultNotifications' : [ u'Thumbnail Done',
                                             u'Post flight script complete',
                                             u'Batch Done' ] }

    def init(self):
        self = super(ThumbscrewAppDelegate, self).init()

        Growl.GrowlApplicationBridge.setGrowlDelegate_(self)

        self.awake = False
        self.bundleIdent = NSBundle.mainBundle().bundleIdentifier()

        self.thumbnailer = None
        self.resizer = None

        self.queue = None
        self.subprocsReady = False

        self.activity_output_controller = None
        self.activity_output = []

        self.ws = NSWorkspace.sharedWorkspace()
        NSUserDefaults.resetStandardUserDefaults()
        self.sud = NSUserDefaults.standardUserDefaults()
        pdomain = self.sud.persistentDomainForName_(self.bundleIdent)
        if pdomain is None:
            self.sud.setPersistentDomain_forName_(
                NSDictionary.dictionaryWithContentsOfFile_(
                    NSBundle.mainBundle().pathForResource_ofType_(
                        self.bundleIdent, u"plist")),
                self.bundleIdent)
            self.sud.synchronize()

        # We instantiate the PreferenceController, because we want to
        # always link our prefs
        self.preference_controller = PreferenceController.PreferenceController.alloc().initWithApplication_(self)

        return self

    def reloadPrefs_(self, sender):
        self.preference_controller.reloadPrefs()

    def setUpDropZone_(self, sender):
        self.progress_bar.stopAnimation_(self)
        self.progress_bar.setIndeterminate_(False)
        self.progress_name.setStringValue_(u'')
        self.drop_zone.can_accept_drags = True
        self.drop_zone.resetImage()
        self.drop_zone.setEditable_(False)
        #self.drop_zone.setImageFrameStyle_(NSImageFrameGrayBezel)
        self.drop_zone.registerForDraggedTypes_([NSFilenamesPboardType,
                                                 NSURLPboardType,
                                                 ])
        self.subprocsReady = True
        if self.queue:
            self.sendBatch()
            self.queue = None


AppHelper.runEventLoop(argv=[])
